@@ -38,13 +38,13 @@ from BorutaShap import BorutaShap
 from xgboost import XGBClassifier, DMatrix, train
 from optuna.integration import TFKerasPruningCallback
 optuna.logging.set_verbosity(optuna.logging.WARNING)
-from MicroLIA.data_augmentation import augmentation, resize
-from MicroLIA import data_processing, cnn_model
+from pyBIA.data_augmentation import augmentation, resize
+from pyBIA import data_processing, cnn_model
 
 
 class objective_cnn(object):
     """
-    Optimization objective function for MicroLIA's convolutional neural network. 
+    Optimization objective function for pyBIA's convolutional neural network. 
     This is passed through the hyper_opt() function when optimizing with
     Optuna. The Optuna software for hyperparameter optimization was published in 
     2019 by Akiba et al. Paper: https://arxiv.org/abs/1907.10902
@@ -85,7 +85,7 @@ class objective_cnn(object):
     Args:
         positive_class (ndarray): The samples for the first class should be passed, which will automatically 
             be assigned the positive label '1'.
-        negative_class (ndarray, str): The samples for the second class should be passed, which will automatically 
+        negative_class (ndarray): The samples for the second class should be passed, which will automatically 
             be assigned the negative label '0'.
         img_num_channels (int): The number of filters. Defaults to 1.
         normalize (bool, optional): If True the data will be min-max normalized using the 
@@ -109,9 +109,9 @@ class objective_cnn(object):
         opt_model (bool): If True, the architecture parameters will be optimized. Defaults to True.
         opt_aug (bool): If True, the augmentation procedure will be optimized. Defaults to False.
         batch_min (int): The minimum number of augmentations to perform per image on the positive class, only applicable 
-        if opt_aug=True. Defaults to 2.
+            if opt_aug=True. Defaults to 2.
         batch_max (int): The maximum number of augmentations to perform per image on the positive class, only applicable 
-        if opt_aug=True. Defaults to 25.
+            if opt_aug=True. Defaults to 25.
         image_size_min (int): The minimum image size to assess, only applicable if opt_aug=True. Defaults to 50.
         image_size_max (int): The maximum image size to assess, only applicable if opt_aug=True. Defaults to 100.
         opt_max_min_pix (int, optional): The minimum max pixel value to use when tuning the normalization procedure, 
@@ -127,16 +127,16 @@ class objective_cnn(object):
             this is set to a value greater than one, overlap may occur. 
         verbose (int): Controls the amount of output printed during the training process. A value of 0 is for silent mode, 
             a value of 1 is used for progress bar mode, and 2 for one line per epoch mode. Defaults to 1.
-        opt_cv (int): Cross-validations to perform when assesing the performance at each
+        opt_cv (int): Cross-validations to perform when assessing the performance at each
             hyperparameter optimization trial. For example, if cv=3, then each optimization trial
             will be assessed according to the 3-fold cross validation accuracy. Defaults to 10.
             NOTE: The higher this number, the longer the optimization will take.
         balance (bool, optional): This will determine whether the two classes
             are kept the same size during optimization, applicable if tuning the augmentation
             parameters. Defaults to True.
-        limit_search (bool): If False, the AlexNet layer paremters will be tuned, including the number of filters to learn
-            as well as the convolution sizes and strides. Defaults to False due to memory allocation issues when handling
-            lots of tunable parameters.
+        limit_search (bool): Whether to expand the hyperparameter search space. If clf='cnn', this will 
+            include the tuning of the individual layer parameters such as the number of filter, size & stride.
+            Defaults to True due to memory allocation issues when handling lots of tunable parameters.
         batch_size_min (int): The minimum batch size to use during training. Should be multiples of 16 for optimal hardware use?? Defaults to 16.
         batch_size_max (int): The Maximum batch size to use during training. Should be multiples of 16 for optimal hardware use?? Defaults to 64.
         train_acc_threshold (float, optional): A value between 0 and 1 that designates the training binary_accuracy threshold that should
@@ -158,7 +158,7 @@ class objective_cnn(object):
         val_positive=None, val_negative=None, test_positive=None, test_negative=None, train_epochs=25, patience=20, 
         opt_model=True, opt_aug=False, batch_min=2, batch_max=25, image_size_min=50, image_size_max=100, 
         balance=True, opt_max_min_pix=None, opt_max_max_pix=None, metric='loss', average=True, shift=10,
-        mask_size=None, num_masks=None, opt_cv=None, verbose=0, train_acc_threshold=None, 
+        mask_size=None, num_masks=None, opt_cv=None, verbose=0, train_acc_threshold=None, smote_sampling=0,
         limit_search=True, batch_size_min=16, batch_size_max=64, monitor1=None, monitor2=None, monitor1_thresh=None, monitor2_thresh=None):
 
         self.positive_class = positive_class
@@ -199,6 +199,7 @@ class objective_cnn(object):
         self.monitor2 = monitor2
         self.monitor1_thresh = monitor1_thresh
         self.monitor2_thresh = monitor2_thresh
+        self.smote_sampling = smote_sampling
 
         if 'all' not in self.metric and 'loss' not in self.metric and 'f1_score' not in self.metric and 'binary_accuracy' not in self.metric:
             raise ValueError("Invalid metric input, options are: 'loss', 'binary_accuracy', 'f1_score', or 'all', and the validation equivalents (add val_ at the beginning).")
@@ -432,7 +433,7 @@ class objective_cnn(object):
                     epochs=self.train_epochs, batch_size=batch_size, lr=lr, momentum=momentum, decay=decay, nesterov=nesterov, 
                     loss=loss, activation_conv=activation_conv, activation_dense=activation_dense, model_reg=model_reg, 
                     conv_init=conv_init, dense_init=dense_init, optimizer=optimizer, early_stop_callback=callbacks, checkpoint=False, verbose=self.verbose, 
-                    pooling_1=pooling_1, pooling_2=pooling_2, pooling_3=pooling_3)
+                    pooling_1=pooling_1, pooling_2=pooling_2, pooling_3=pooling_3, smote_sampling=self.smote_sampling)
             else:
                 ### Filter and Layer Characterstics ###
                 filter_1 = trial.suggest_int('filter_1', 12, 240, step=12)
@@ -481,7 +482,8 @@ class objective_cnn(object):
                     filter_3=filter_3, filter_size_3=filter_size_3, strides_3=strides_3, filter_4=filter_4, 
                     filter_size_4=filter_size_4, strides_4=strides_4, filter_5=filter_5, filter_size_5=filter_size_5, 
                     strides_5=strides_5, dense_neurons_1=dense_neurons_1, dense_neurons_2=dense_neurons_2, 
-                    dropout_1=dropout_1, dropout_2=dropout_2, early_stop_callback=callbacks, checkpoint=False, verbose=self.verbose)
+                    dropout_1=dropout_1, dropout_2=dropout_2, smote_sampling=self.smote_sampling,
+                    early_stop_callback=callbacks, checkpoint=False, verbose=self.verbose)
         else:
             if self.opt_cv is not None:
                 print()
@@ -498,7 +500,7 @@ class objective_cnn(object):
             model, history = cnn_model.AlexNet(class_1, class_2, img_num_channels=self.img_num_channels, normalize=self.normalize, 
                 min_pixel=min_pix, max_pixel=max_pix, val_positive=val_class_1, val_negative=val_class_2, epochs=self.train_epochs, 
                 batch_size=batch_size, lr=lr, momentum=momentum, decay=decay, nesterov=nesterov, early_stop_callback=callbacks, 
-                checkpoint=False, verbose=self.verbose, optimizer=optimizer)
+                checkpoint=False, verbose=self.verbose, optimizer=optimizer, smote_sampling=self.smote_sampling)
 
         #If the patience is reached, return nan#
         if len(history.history['loss']) != self.train_epochs:
@@ -517,7 +519,6 @@ class objective_cnn(object):
                     print()
                     return (len(history.history['loss']) * 0.001) - 999.0
 
-
             models, histories = [], []
             models.append(model), histories.append(history)
 
@@ -650,7 +651,7 @@ class objective_cnn(object):
                     model, history = cnn_model.AlexNet(class_1, class_2, img_num_channels=self.img_num_channels, normalize=self.normalize, 
                         min_pixel=min_pix, max_pixel=max_pix, val_positive=val_class_1, val_negative=val_class_2, epochs=self.train_epochs, 
                         batch_size=batch_size, lr=lr, momentum=momentum, decay=decay, nesterov=nesterov, early_stop_callback=callbacks, 
-                        conv_init=conv_init, dense_init=dense_init, checkpoint=False, verbose=self.verbose, optimizer=optimizer)
+                        conv_init=conv_init, dense_init=dense_init, checkpoint=False, verbose=self.verbose, optimizer=optimizer, smote_sampling=self.smote_sampling)
                 else:
                     if self.limit_search:
                         model, history = cnn_model.AlexNet(class_1, class_2, img_num_channels=self.img_num_channels, 
@@ -658,7 +659,7 @@ class objective_cnn(object):
                             epochs=self.train_epochs, batch_size=batch_size, lr=lr, momentum=momentum, decay=decay, nesterov=nesterov, 
                             loss=loss, activation_conv=activation_conv, activation_dense=activation_dense, model_reg=model_reg, 
                             conv_init=conv_init, dense_init=dense_init, optimizer=optimizer, early_stop_callback=callbacks, checkpoint=False, verbose=self.verbose, 
-                            pooling_1=pooling_1, pooling_2=pooling_2, pooling_3=pooling_3)
+                            pooling_1=pooling_1, pooling_2=pooling_2, pooling_3=pooling_3, smote_sampling=self.smote_sampling)
                     else:
                         model, history = cnn_model.AlexNet(class_1, class_2, img_num_channels=self.img_num_channels, 
                             normalize=self.normalize, min_pixel=min_pix, max_pixel=max_pix, val_positive=val_class_1, val_negative=val_class_2, 
@@ -671,7 +672,7 @@ class objective_cnn(object):
                             filter_3=filter_3, filter_size_3=filter_size_3, strides_3=strides_3, filter_4=filter_4, 
                             filter_size_4=filter_size_4, strides_4=strides_4, filter_5=filter_5, filter_size_5=filter_size_5, 
                             strides_5=strides_5, dense_neurons_1=dense_neurons_1, dense_neurons_2=dense_neurons_2, 
-                            dropout_1=dropout_1, dropout_2=dropout_2, early_stop_callback=callbacks, checkpoint=False, verbose=self.verbose)
+                            dropout_1=dropout_1, dropout_2=dropout_2, smote_sampling=self.smote_sampling, early_stop_callback=callbacks, checkpoint=False, verbose=self.verbose)
 
                 #If the trial is pruned or the patience is reached, return nan#
                 if len(history.history['loss']) != self.train_epochs:
@@ -1106,10 +1107,10 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, return_study=True, balance=Tr
     normalize=True, min_pixel=0, max_pixel=100, val_X=None, val_Y=None, train_epochs=25, patience=0, metric='loss', 
     limit_search=True, opt_model=True, opt_aug=False, batch_min=2, batch_max=25, image_size_min=50, image_size_max=100,
     opt_max_min_pix=None, opt_max_max_pix=None, opt_cv=10, test_size=None, average=True, test_positive=None, test_negative=None, shift=10, 
-    mask_size=None, num_masks=None, verbose=1, train_acc_threshold=None, monitor1=None, monitor2=None, monitor1_thresh=None, monitor2_thresh=None):
+    mask_size=None, num_masks=None, verbose=1, train_acc_threshold=None, monitor1=None, monitor2=None, monitor1_thresh=None, monitor2_thresh=None,
+    smote_sampling=0):
     """
-    Optimizes hyperparameters using a k-fold cross validation splitting strategy, unless a CNN
-    is being optimized, in which case no cross-validation is performed during trial assesment.
+    Optimizes hyperparameters using a k-fold cross validation splitting strategy.
 
     **IMPORTANT** In the case of CNN optimization, data_x and data_y are not the standard
     data plus labels -- if optimizing a CNN the samples for the first class should be passed
@@ -1219,7 +1220,9 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, return_study=True, balance=Tr
             this is set to a value greater than one, overlap may occur. 
         verbose (int): Controls the amount of output printed during the training process. A value of 0 is for silent mode, 
             a value of 1 is used for progress bar mode, and 2 for one line per epoch mode. Defaults to 1.
-
+        smote_sampling (float): The smote_sampling parameter is used in the SMOTE algorithm to specify the desired 
+            ratio of the minority class to the majority class. Defaults to 0 which disables the procedure.
+        
     The folllowing arguments can be used to set early-stopping callbacks. These can be used to terminate trials that exceed
     pre-determined thresholds, which may be indicative of an overfit model.
 
@@ -1378,7 +1381,7 @@ def hyper_opt(data_x, data_y, clf='rf', n_iter=25, return_study=True, balance=Tr
             val_positive=val_X, val_negative=val_Y, train_epochs=train_epochs, patience=patience, metric=metric, average=average, test_positive=test_positive, test_negative=test_negative,
             opt_model=opt_model, opt_aug=opt_aug, batch_min=batch_min, batch_max=batch_max, image_size_min=image_size_min, image_size_max=image_size_max, balance=balance, 
             opt_max_min_pix=opt_max_min_pix, opt_max_max_pix=opt_max_max_pix, shift=shift, opt_cv=opt_cv, mask_size=mask_size, num_masks=num_masks, train_acc_threshold=train_acc_threshold,
-            verbose=verbose, limit_search=limit_search, monitor1=monitor1, monitor2=monitor2, monitor1_thresh=monitor1_thresh, monitor2_thresh=monitor2_thresh)
+            verbose=verbose, limit_search=limit_search, monitor1=monitor1, monitor2=monitor2, monitor1_thresh=monitor1_thresh, monitor2_thresh=monitor2_thresh, smote_sampling=smote_sampling)
         study.optimize(objective, n_trials=n_iter, show_progress_bar=True, gc_after_trial=True)#, n_jobs=1)
         params = study.best_trial.params
 
